from skrl.trainers.jax.base import Trainer, generate_equally_spaced_scopes  # isort:skip

from skrl.trainers.jax.manual import ManualTrainer
from skrl.trainers.jax.sequential import SequentialTrainer
